/*
 * Copyright (c) 2007 Pentaho Corporation.  All rights reserved.
 * This software was developed by Pentaho Corporation and is provided under the terms
 * of the GNU Lesser General Public License, Version 2.1. You may not use
 * this file except in compliance with the license. If you need a copy of the license,
 * please go to http://www.gnu.org/licenses/lgpl-2.1.txt. The Original Code is Pentaho
 * Data Integration.  The Initial Developer is Pentaho Corporation.
 *
 * Software distributed under the GNU Lesser Public License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or  implied. Please refer to
 * the license for the specific language governing your rights and limitations.
 */

package org.pentaho.di.core.database;

import org.pentaho.di.core.Const;
import org.pentaho.di.core.row.ValueMetaInterface;

/**
 * Contains Hypersonic specific information through static final members
 *
 * @author Matt
 * @since 11-mrt-2005
 */

public class H2DatabaseMeta extends BaseDatabaseMeta implements DatabaseInterface {
    public int[] getAccessTypeList() {
        return new int[]{DatabaseMeta.TYPE_ACCESS_NATIVE, DatabaseMeta.TYPE_ACCESS_ODBC, DatabaseMeta.TYPE_ACCESS_JNDI};
    }

    /**
     * @see DatabaseInterface#getNotFoundTK(boolean)
     */
    public int getNotFoundTK(boolean use_autoinc) {
        return super.getNotFoundTK(use_autoinc);
    }

    public String getDriverClass() {
        if (getAccessType() == DatabaseMeta.TYPE_ACCESS_NATIVE) {
            return "org.h2.Driver";
        } else {
            return "sun.jdbc.odbc.JdbcOdbcDriver"; // always ODBC!
        }

    }

    public int getDefaultDatabasePort() {
        if (getAccessType() == DatabaseMeta.TYPE_ACCESS_NATIVE) return 8082;
        return -1;
    }

    public String getURL(String hostname, String port, String databaseName) {
        if (getAccessType() == DatabaseMeta.TYPE_ACCESS_NATIVE) {
            // If the database is an in-memory DB or if there is no valid port and hostname, go embedded
            //
            if ((databaseName != null && databaseName.startsWith("mem:")) ||
                    ((Const.isEmpty(port) || "-1".equals(port)) && Const.isEmpty(hostname))) {
                return "jdbc:h2:" + databaseName;
            } else {
                // Connect over TCP/IP
                //
                return "jdbc:h2:tcp://" + hostname + ":" + port + "/" + databaseName;
            }
        } else {
            return "jdbc:odbc:" + databaseName;
        }
    }

    /**
     * Checks whether or not the command setFetchSize() is supported by the JDBC driver...
     *
     * @return true is setFetchSize() is supported!
     */
    public boolean isFetchSizeSupported() {
        return true;
    }

    /**
     * @see DatabaseInterface#getSchemaTableCombination(String, String)
     */
    public String getSchemaTableCombination(String schema_name, String table_part) {
        return getBackwardsCompatibleSchemaTableCombination(schema_name, table_part);
    }

    /**
     * @return true if the database supports bitmap indexes
     */
    public boolean supportsBitmapIndex() {
        return false;
    }

    /**
     * @return true if Kettle can create a repository on this type of database.
     */
    public boolean supportsRepository() {
        return true;
    }

    public boolean supportsAutoInc() {
        return true;
    }

    public boolean supportsGetBlob() {
        return true;
    }

    public boolean supportsSetCharacterStream() {
        return false;
    }

    /**
     * Generates the SQL statement to add a column to the specified table
     * For this generic type, i set it to the most common possibility.
     *
     * @param tablename   The table to add
     * @param v           The column defined as a value
     * @param tk          the name of the technical key field
     * @param use_autoinc whether or not this field uses auto increment
     * @param pk          the name of the primary key field
     * @param semicolon   whether or not to add a semi-colon behind the statement.
     * @return the SQL statement to add a column to the specified table
     */
    public String getAddColumnStatement(String tablename, ValueMetaInterface v, String tk, boolean use_autoinc, String pk, boolean semicolon) {
        return "ALTER TABLE " + tablename + " ADD " + getFieldDefinition(v, tk, pk, use_autoinc, true, false);
    }

    /**
     * Generates the SQL statement to modify a column in the specified table
     *
     * @param tablename   The table to add
     * @param v           The column defined as a value
     * @param tk          the name of the technical key field
     * @param use_autoinc whether or not this field uses auto increment
     * @param pk          the name of the primary key field
     * @param semicolon   whether or not to add a semi-colon behind the statement.
     * @return the SQL statement to modify a column in the specified table
     */
    public String getModifyColumnStatement(String tablename, ValueMetaInterface v, String tk, boolean use_autoinc, String pk, boolean semicolon) {
        return "ALTER TABLE " + tablename + " ALTER " + getFieldDefinition(v, tk, pk, use_autoinc, true, false);
    }

    public String getFieldDefinition(ValueMetaInterface v, String tk, String pk, boolean use_autoinc, boolean add_fieldname, boolean add_cr) {
        String retval = "";

        String fieldname = v.getName();
        int length = v.getLength();
        int precision = v.getPrecision();

        if (add_fieldname) retval += fieldname + " ";

        int type = v.getType();
        switch (type) {
            case ValueMetaInterface.TYPE_DATE:
                retval += "TIMESTAMP";
                break;
            case ValueMetaInterface.TYPE_BOOLEAN:
                if (supportsBooleanDataType()) {
                    retval += "BOOLEAN";
                } else {
                    retval += "CHAR(1)";
                }
                break;
            case ValueMetaInterface.TYPE_NUMBER:
            case ValueMetaInterface.TYPE_INTEGER:
            case ValueMetaInterface.TYPE_BIGNUMBER:
                if (fieldname.equalsIgnoreCase(tk) || // Technical key
                        fieldname.equalsIgnoreCase(pk)    // Primary key
                ) {
                    retval += "IDENTITY";
                } else {
                    if (length > 0) {
                        if (precision > 0 || length > 18) {
                            retval += "DECIMAL(" + length + ", " + precision + ")";
                        } else {
                            if (length > 9) {
                                retval += "BIGINT";
                            } else {
                                if (length < 5) {
                                    if (length < 3) {
                                        retval += "TINYINT";
                                    } else {
                                        retval += "SMALLINT";
                                    }
                                } else {
                                    retval += "INTEGER";
                                }
                            }
                        }

                    } else {
                        retval += "DOUBLE";
                    }
                }
                break;
            case ValueMetaInterface.TYPE_STRING:
                if (length >= DatabaseMeta.CLOB_LENGTH) {
                    retval += "TEXT";
                } else {
                    retval += "VARCHAR";
                    if (length > 0) {
                        retval += "(" + length;
                    } else {
                        // http://www.h2database.com/html/datatypes.html#varchar_type
                        retval += "(" + Integer.MAX_VALUE;
                    }
                    retval += ")";
                }
                break;
            case ValueMetaInterface.TYPE_BINARY:
                retval += "BLOB";
                break;
            default:
                retval += "UNKNOWN";
                break;
        }

        if (add_cr) retval += Const.CR;

        return retval;
    }

    public String[] getReservedWords() {
        return new String[]
                {
                        "CURRENT_TIMESTAMP", "CURRENT_TIME", "CURRENT_DATE", "CROSS", "DISTINCT", "EXCEPT", "EXISTS", "FROM", "FOR", "FALSE",
                        "FULL", "GROUP", "HAVING", "INNER", "INTERSECT", "IS", "JOIN", "LIKE", "MINUS", "NATURAL", "NOT", "NULL", "ON",
                        "ORDER", "PRIMARY", "ROWNUM", "SELECT", "SYSDATE", "SYSTIME", "SYSTIMESTAMP", "TODAY", "TRUE", "UNION", "WHERE",
                };
    }

    public String[] getUsedLibraries() {
        return new String[]{"h2.jar"};
    }

    /**
     * Most databases allow you to retrieve result metadata by preparing a SELECT statement.
     *
     * @return true if the database supports retrieval of query metadata from a prepared statement.  False if the query needs to be executed first.
     */
    public boolean supportsPreparedStatementMetadataRetrieval() {
        return false;
    }

}
